@@ -22,6 +22,9 @@ CONFIG_DIR = Path(user_config_dir(APP_NAME, APP_AUTHOR))
 DATA_DIR = Path(user_data_dir(APP_NAME, APP_AUTHOR))
 CACHE_DIR = Path(user_cache_dir(APP_NAME, APP_AUTHOR))
 
+MIN_WIDTH = 1280
+MIN_HEIGHT = 900
+
 
 # Sentinel for tracking unset fields
 NOT_PRESENT = object()
@@ -84,9 +87,12 @@ def guess_install_dir():
 class ConfigFile:
     def __init__(self, config_path: Path):
         self.config_path = config_path
+        self.dirty = False
 
         self.install_dir = None
         self.playlunky_version = None
+        self.playlunky_console = False
+        self.geometry = None
 
     @classmethod
     def from_path(cls, config_path: Path):
@@ -95,7 +101,15 @@ class ConfigFile:
 
         if config_path.exists():
             with config_path.open("r") as config_file:
-                config_data = json.load(config_file)
+                try:
+                    config_data = json.load(config_file)
+                except Exception as err:  # pylint: disable=broad-except
+                    logger.critical(
+                        "Failed to read config file: %s. Creating new one from defaults.",
+                        err,
+                    )
+                    needs_save = True
+                    config_data = {}
         else:
             config_data = {}
 
@@ -108,8 +122,12 @@ class ConfigFile:
             install_dir = Path(install_dir)
         obj.install_dir = install_dir
 
-        # Initialize playlunky-version
+        # Initialize playlunky config
         obj.playlunky_version = config_data.get("playlunky-version")
+        obj.playlunky_console = config_data.get("playlunky-console", False)
+
+        # Initialize geometry
+        obj.geometry = config_data.get("geometry", f"{MIN_WIDTH}x{MIN_HEIGHT}")
 
         if needs_save:
             obj.save()
@@ -126,6 +144,9 @@ class ConfigFile:
 
         if self.playlunky_version is not None:
             out["playlunky-version"] = self.playlunky_version
+        out["playlunky-console"] = self.playlunky_console
+
+        out["geometry"] = self.geometry
 
         return out
 
@@ -133,6 +154,8 @@ class ConfigFile:
         return self.config_path.with_suffix(f"{self.config_path.suffix}.tmp")
 
     def save(self):
+        self.dirty = False
+
         # Make a temporary file so we can do an atomic replace
         # in case something crashes while writing out the config.
         tmp_path = self._get_tmp_path()
