@@ -1,14 +1,20 @@
+# pylint: disable=too-many-lines
+
 import json
 import logging
+import shutil
 import subprocess
 import threading
+import time
 import tkinter as tk
+import webbrowser
 import zipfile
 from io import BytesIO
 from pathlib import Path
 from shutil import copyfile
+from tkinter import PhotoImage
 from tkinter import font as tk_font
-from tkinter import messagebox, ttk
+from tkinter import ttk
 
 import requests
 from PIL import Image, ImageTk
@@ -16,7 +22,7 @@ from PIL import Image, ImageTk
 from modlunky2.config import CACHE_DIR, DATA_DIR
 from modlunky2.constants import BASE_DIR
 from modlunky2.ui.play.config import PlaylunkyConfig
-from modlunky2.ui.widgets import ScrollableFrame, Tab
+from modlunky2.ui.widgets import ScrollableFrame, Tab, ToolTip
 
 logger = logging.getLogger("modlunky2")
 
@@ -25,9 +31,10 @@ PLAYLUNKY_RELEASES_URL = "https://api.github.com/repos/spelunky-fyi/Playlunky/re
 PLAYLUNKY_RELEASES_PATH = CACHE_DIR / "playlunky-releases.json"
 PLAYLUNKY_DATA_DIR = DATA_DIR / "playlunky"
 
+SPEL2_DLL = "spel2.dll"
 PLAYLUNKY_DLL = "playlunky64.dll"
 PLAYLUNKY_EXE = "playlunky_launcher.exe"
-PLAYLUNKY_FILES = [PLAYLUNKY_DLL, PLAYLUNKY_EXE]
+PLAYLUNKY_FILES = [SPEL2_DLL, PLAYLUNKY_DLL, PLAYLUNKY_EXE]
 
 
 class Entry(tk.Entry):
@@ -96,7 +103,7 @@ def cache_playlunky_releases(call):
     call("play:cache_releases_updated")
 
 
-def download_playlunky_release(call, tag, download_url):
+def download_playlunky_release(call, tag, download_url, launch):
 
     dest_path = PLAYLUNKY_DATA_DIR / tag
     if not dest_path.exists():
@@ -116,7 +123,8 @@ def download_playlunky_release(call, tag, download_url):
     for member in playlunky_zip.infolist():
         if member.filename in PLAYLUNKY_FILES:
             playlunky_zip.extract(member, dest_path)
-    call("play:download_finished")
+
+    call("play:download_finished", launch=launch)
 
 
 class DownloadFrame(tk.Frame):
@@ -148,7 +156,7 @@ class DownloadFrame(tk.Frame):
         self.progress_bar = ttk.Progressbar(self)
         self.progress_bar.grid(row=3, column=0, pady=5, padx=5, sticky="we")
 
-    def download(self):
+    def download(self, launch=False):
         tag = self.parent.selected_var.get()
         release = self.parent.available_releases.get(tag)
         if release is None:
@@ -164,20 +172,25 @@ class DownloadFrame(tk.Frame):
         self.progress_bar["maximum"] = asset["size"]
         self.button["state"] = tk.DISABLED
         self.parent.selected_dropdown["state"] = tk.DISABLED
+        self.parent.parent.disable_button()
         self.task_manager.call(
             "play:start_download",
             tag=tag,
             download_url=asset["browser_download_url"],
+            launch=launch,
         )
 
     def on_download_progress(self, amount_downloaded):
         self.progress_bar["value"] = amount_downloaded
 
-    def on_download_finished(self):
-        self.parent.parent.enable_button()
+    def on_download_finished(self, launch=False):
         self.progress_bar["value"] = 0
-        self.button["state"] = tk.NORMAL
-        self.parent.render()
+        if launch:
+            self.parent.parent.play()
+        else:
+            self.parent.parent.enable_button()
+            self.button["state"] = tk.NORMAL
+            self.parent.render()
 
 
 def uninstall_playlunky_release(call, tag):
@@ -228,6 +241,7 @@ class UninstallFrame(tk.Frame):
     def on_uninstall_finished(self):
         self.button["state"] = tk.NORMAL
         self.parent.render()
+        self.parent.parent.enable_button()
 
 
 def is_installed(tag):
@@ -235,7 +249,7 @@ def is_installed(tag):
 
 
 class VersionFrame(tk.LabelFrame):
-    CACHE_RELEASES_INTERVAL = 1000 * 10 * 60
+    CACHE_RELEASES_INTERVAL = 1000 * 30 * 60
 
     def __init__(self, parent, config, task_manager):
         super().__init__(parent, text="Version")
@@ -298,7 +312,25 @@ class VersionFrame(tk.LabelFrame):
         return available_releases
 
     def cache_releases(self):
+        next_run = self.CACHE_RELEASES_INTERVAL
+        if PLAYLUNKY_RELEASES_PATH.exists():
+            mtime = int(PLAYLUNKY_RELEASES_PATH.stat().st_mtime) * 1000
+            now = int(time.time()) * 1000
+            delta = now - mtime
+            if delta < self.CACHE_RELEASES_INTERVAL - 1000:
+                next_run = self.CACHE_RELEASES_INTERVAL - delta
+                logger.debug(
+                    "Playlunky releases were retrieved too recently. Retrying in %s seconds",
+                    next_run / 1000,
+                )
+                self.after(next_run, self.cache_releases)
+                return
+
         self.task_manager.call("play:cache_releases")
+        logger.debug(
+            "Scheduling next run for caching releases in %s seconds",
+            self.CACHE_RELEASES_INTERVAL / 1000,
+        )
         self.after(self.CACHE_RELEASES_INTERVAL, self.cache_releases)
 
     def release_selected(self, value):
@@ -334,6 +366,11 @@ class VersionFrame(tk.LabelFrame):
         selected_version = self.config.config_file.playlunky_version
         if selected_version:
             self.selected_var.set(selected_version)
+        else:
+            selected_version = available_releases[0]
+            self.config.config_file.playlunky_version = selected_version
+            self.config.config_file.save()
+            self.selected_var.set(selected_version)
 
         for release in self.available_releases:
             if release in installed_releases:
@@ -343,19 +380,20 @@ class VersionFrame(tk.LabelFrame):
 
         if selected_version in installed_releases:
             self.show_uninstall_frame()
-            self.parent.enable_button()
         else:
             self.show_download_frame()
-            self.parent.disable_button()
+
+        self.parent.enable_button()
 
     def on_cache_releases_updated(self):
         self.render()
 
 
 class OptionsFrame(tk.LabelFrame):
-    def __init__(self, parent):
+    def __init__(self, parent, config):
         super().__init__(parent, text="Options")
         self.parent = parent
+        self.config = config
         self.columnconfigure(0, weight=1)
 
         self.random_char_var = tk.BooleanVar()
@@ -385,6 +423,30 @@ class OptionsFrame(tk.LabelFrame):
         )
         self.cache_decoded_audio_checkbox.grid(row=2, column=0, sticky="w")
 
+        self.enable_developer_mode_var = tk.BooleanVar()
+        self.enable_developer_mode_checkbox = tk.Checkbutton(
+            self,
+            text="Enable Developer Mode",
+            variable=self.enable_developer_mode_var,
+            compound="left",
+        )
+        self.enable_developer_mode_checkbox.grid(row=3, column=0, sticky="w")
+
+        self.enable_console_var = tk.BooleanVar()
+        self.enable_console_var.set(self.config.config_file.playlunky_console)
+        self.enable_console_checkbox = tk.Checkbutton(
+            self,
+            text="Enable Console",
+            variable=self.enable_console_var,
+            compound="left",
+            command=self.handle_console_checkbutton,
+        )
+        self.enable_console_checkbox.grid(row=4, column=0, sticky="w")
+
+    def handle_console_checkbutton(self):
+        self.config.config_file.playlunky_console = self.enable_console_var.get()
+        self.config.config_file.save()
+
 
 class FiltersFrame(tk.LabelFrame):
     def __init__(self, parent):
@@ -424,15 +486,67 @@ class FiltersFrame(tk.LabelFrame):
             self.parent.master.render_packs()
 
 
-class InstallModFrame(tk.LabelFrame):
-    def __init__(self, parent, *args, **kwargs):
-        super().__init__(parent, text="Install Mod", *args, **kwargs)
+class ControlsFrame(tk.LabelFrame):
+    def __init__(self, parent, config, *args, **kwargs):
+        super().__init__(parent, text="Stuff & Things", *args, **kwargs)
         self.parent = parent
-        self.columnconfigure(0, weight=1)
-        self.rowconfigure(0, weight=1)
+        self.config = config
 
-        self.button = ttk.Button(self, text="Browse", command=self.install_mod)
-        self.button.grid(row=0, column=0, pady=5, padx=10, sticky="nswe")
+        self.columnconfigure(0, weight=1)
+
+        # self.install_button = ttk.Button(
+        #    self, text="Install Mod", command=self.install_mod
+        # )
+        # self.install_button.grid(row=0, column=0, pady=3, padx=10, sticky="nswe")
+        # ToolTip(self.install_button, "Browse for a mod file to install.")
+
+        self.refresh_button = ttk.Button(
+            self, text="Refresh Mods", command=self.refresh_mods
+        )
+        self.refresh_button.grid(row=1, column=0, pady=3, padx=10, sticky="nswe")
+        ToolTip(
+            self.refresh_button,
+            (
+                "If you've made any changes in the Packs directory\n"
+                "that you want updated in the mod list."
+            ),
+        )
+
+        self.open_packs_button = ttk.Button(
+            self, text="Open Packs Directory", command=self.open_packs
+        )
+        self.open_packs_button.grid(row=2, column=0, pady=3, padx=10, sticky="nswe")
+        ToolTip(self.open_packs_button, ("Open the directory where Packs are saved"))
+
+        self.guide_button = ttk.Button(self, text="User Guide", command=self.guide)
+        self.guide_button.grid(row=3, column=0, pady=3, padx=10, sticky="nswe")
+        ToolTip(self.guide_button, ("Open the User Guide"))
+
+        self.update_releases_button = ttk.Button(
+            self, text="Update Releases", command=self.update_releases
+        )
+        self.update_releases_button.grid(
+            row=4, column=0, pady=3, padx=10, sticky="nswe"
+        )
+        ToolTip(
+            self.update_releases_button,
+            (
+                "If you want to check for a new version of Playlunky\n"
+                "you can force an update with this button."
+            ),
+        )
+
+        self.clear_cache_button = ttk.Button(
+            self, text="Clear Cache", command=self.clear_cache
+        )
+        self.clear_cache_button.grid(row=5, column=0, pady=3, padx=10, sticky="nswe")
+        ToolTip(
+            self.clear_cache_button,
+            (
+                "Remove Playlunky cache. This could be helpful\n"
+                "if things aren't working as expected."
+            ),
+        )
 
     def install_mod(self):
         tk.messagebox.showinfo(
@@ -440,6 +554,44 @@ class InstallModFrame(tk.LabelFrame):
             "This feature hasn't been implemented yet, but it will be soon!",
         )
 
+    def refresh_mods(self):
+        self.parent.on_load()
+
+    def open_packs(self):
+        packs_dir = self.config.install_dir / "Mods/Packs"
+        if not packs_dir.exists():
+            logger.info("Couldn't find Packs directory. Looked in %s", packs_dir)
+            return
+
+        webbrowser.open(f"file://{packs_dir}")
+
+    def guide(self):
+        webbrowser.open_new_tab("https://github.com/spelunky-fyi/Playlunky/wiki")
+
+    def update_releases(self):
+        self.parent.version_frame.task_manager.call("play:cache_releases")
+
+    def clear_cache(self):
+        cache_dir = self.config.install_dir / "Mods/Packs/.db"
+        if not cache_dir.exists():
+            logger.info("No cache directory found to remove. Looked in %s", cache_dir)
+            return
+
+        answer = tk.messagebox.askokcancel(
+            title="Confirmation",
+            message=(
+                "Are you sure you want to remove Playlunky cache?\n"
+                "\n"
+                f"This will remove {cache_dir} and all of its contents."
+            ),
+            icon=tk.messagebox.WARNING,
+        )
+
+        if not answer:
+            return
+
+        shutil.rmtree(cache_dir)
+
 
 class LoadOrderFrame(tk.LabelFrame):
     def __init__(self, parent, *args, **kwargs):
@@ -549,12 +701,15 @@ class LoadOrderFrame(tk.LabelFrame):
         self.render_buttons()
 
 
-def launch_playlunky(_call, install_dir, exe_path):
+def launch_playlunky(_call, install_dir, exe_path, use_console):
     logger.info(
         "Executing Playlunky Launcher with %s", exe_path.relative_to(PLAYLUNKY_DATA_DIR)
     )
     working_dir = exe_path.parent
     cmd = [f"{exe_path}", f"--exe_dir={install_dir}"]
+    if use_console:
+        cmd.append("--console")
+
     proc = subprocess.Popen(cmd, cwd=working_dir)
     proc.communicate()
 
@@ -582,6 +737,7 @@ class PlayTab(Tab):
         self.rowconfigure(3, minsize=60)
         self.columnconfigure(0, weight=1)
         self.columnconfigure(1, minsize=250)
+        self.columnconfigure(2, minsize=250)
 
         self.play_wrapper = tk.Frame(self)
         self.play_wrapper.grid(row=0, column=0, rowspan=3, sticky="nswe")
@@ -591,26 +747,33 @@ class PlayTab(Tab):
         self.filter_frame = FiltersFrame(self.play_wrapper)
         self.filter_frame.grid(row=0, column=0, pady=5, padx=5, sticky="nswe")
 
-        self.install_mod_frame = InstallModFrame(self.play_wrapper)
-        self.install_mod_frame.grid(row=0, column=1, pady=5, padx=5, sticky="nswe")
-
         self.packs_frame = ScrollableFrame(
             self.play_wrapper, text="Select Mods to Play"
         )
         self.packs_frame.rowconfigure(0, weight=1)
         self.packs_frame.columnconfigure(0, weight=1)
+        self.packs_frame.scrollable_frame.columnconfigure(0, weight=1)
         self.packs_frame.grid(
             row=1, column=0, columnspan=2, pady=5, padx=5, sticky="nswe"
         )
 
         self.version_frame = VersionFrame(self, config, task_manager)
-        self.version_frame.grid(row=0, column=1, pady=5, padx=5, sticky="nswe")
+        self.version_frame.grid(
+            row=0, column=1, rowspan=2, pady=5, padx=5, sticky="nswe"
+        )
 
-        self.options_frame = OptionsFrame(self)
-        self.options_frame.grid(row=1, column=1, pady=5, padx=5, sticky="nswe")
+        self.install_mod_frame = ControlsFrame(self, config)
+        self.install_mod_frame.grid(
+            row=2, column=1, rowspan=2, pady=5, padx=5, sticky="nswe"
+        )
+
+        self.options_frame = OptionsFrame(self, config)
+        self.options_frame.grid(
+            row=0, column=2, rowspan=2, pady=5, padx=5, sticky="nswe"
+        )
 
         self.load_order = LoadOrderFrame(self)
-        self.load_order.grid(row=2, column=1, rowspan=2, pady=5, padx=5, sticky="nswe")
+        self.load_order.grid(row=2, column=2, rowspan=2, pady=5, padx=5, sticky="nswe")
 
         self.button_play = ttk.Button(
             self, text="Play!", state=tk.DISABLED, command=self.play
@@ -620,10 +783,16 @@ class PlayTab(Tab):
         self.version_frame.render()
         self.version_frame.cache_releases()
 
-        default_icon_path = BASE_DIR / "static/images/noicon.png"
-        self.default_icon = ImageTk.PhotoImage(Image.open(default_icon_path))
+        icon_path = BASE_DIR / "static/images"
+        self.folder_icon = ImageTk.PhotoImage(
+            Image.open(icon_path / "folder.png").resize((30, 30), Image.ANTIALIAS)
+        )
+        self.trash_icon = ImageTk.PhotoImage(
+            Image.open(icon_path / "trash.png").resize((30, 30), Image.ANTIALIAS)
+        )
 
         self.packs = []
+        self.separators = []
         self.checkboxes = {}
 
         self.on_load()
@@ -657,6 +826,7 @@ class PlayTab(Tab):
         self.options_frame.random_char_var.set(config.random_character_select)
         self.options_frame.loose_audio_var.set(config.enable_loose_audio_files)
         self.options_frame.cache_decoded_audio_var.set(config.cache_decoded_audio_files)
+        self.options_frame.enable_developer_mode_var.set(config.enable_developer_mode)
 
     def write_ini(self):
         path = self.config.install_dir / "playlunky.ini"
@@ -664,6 +834,7 @@ class PlayTab(Tab):
             random_character_select=self.options_frame.random_char_var.get(),
             enable_loose_audio_files=self.options_frame.loose_audio_var.get(),
             cache_decoded_audio_files=self.options_frame.cache_decoded_audio_var.get(),
+            enable_developer_mode=self.options_frame.enable_developer_mode_var.get(),
         )
         with path.open("w") as handle:
             config.write(handle)
@@ -682,8 +853,8 @@ class PlayTab(Tab):
                     select = False
                     line = line[2:]
 
-                var, checkbox = self.checkboxes.get(line, (None, None))
-                if (var, checkbox) == (None, None):
+                var, checkbox, buttons = self.checkboxes.get(line, (None, None, None))
+                if (var, checkbox, buttons) == (None, None, None):
                     continue
 
                 if select:
@@ -713,33 +884,62 @@ class PlayTab(Tab):
     def load_order_path(self):
         return self.config.install_dir / "Mods/Packs/load_order.txt"
 
-    def play(self):
-        self.write_steam_appid()
-        self.write_load_order()
-        self.write_ini()
+    def should_install(self):
+        version = self.config.config_file.playlunky_version
+        if version:
+            msg = (
+                f"You don't currently have version {version} installed.\n\n"
+                "Would you like to install it?"
+            )
+        else:
+            msg = (
+                "You don't have any version of Playlunky selected.\n\n"
+                "Would you like to install and run the latest?"
+            )
 
+        answer = tk.messagebox.askokcancel(
+            title="Install?",
+            message=msg,
+            icon=tk.messagebox.INFO,
+        )
+
+        return answer
+
+    def play(self):
         exe_path = (
             PLAYLUNKY_DATA_DIR
             / self.config.config_file.playlunky_version
             / PLAYLUNKY_EXE
         )
+        self.disable_button()
+
         if not exe_path.exists():
-            logger.critical("No playlunky launcher found at %s", exe_path)
+            should_install = self.should_install()
+            if should_install:
+                self.version_frame.download_frame.download(launch=True)
+            else:
+                logger.critical("Can't run without an installed version of Playlunky")
+                self.enable_button()
             return
 
+        self.write_steam_appid()
+        self.write_load_order()
+        self.write_ini()
+
         self.version_frame.selected_dropdown["state"] = tk.DISABLED
         self.version_frame.uninstall_frame.button["state"] = tk.DISABLED
-        self.disable_button()
         self.task_manager.call(
             "play:launch_playlunky",
             install_dir=self.config.install_dir,
             exe_path=exe_path,
+            use_console=self.config.config_file.playlunky_console,
         )
 
     def playlunky_closed(self):
         self.version_frame.selected_dropdown["state"] = tk.NORMAL
         self.version_frame.uninstall_frame.button["state"] = tk.NORMAL
         self.enable_button()
+        self.version_frame.render()
 
     @staticmethod
     def diff_packs(before, after):
@@ -761,6 +961,10 @@ class PlayTab(Tab):
 
             if path.is_file and path.suffix.lower() == ".zip":
                 path = path.with_suffix("")
+                # If There's a version that's extracted with the same name
+                # skip the zipped version
+                if path.exists():
+                    continue
 
             packs.append(str(path.relative_to(self.config.install_dir / "Mods/Packs")))
         return packs
@@ -772,12 +976,23 @@ class PlayTab(Tab):
 
         filter_selected = self.filter_frame.selected_var.get()
 
+        for sep in self.separators:
+            sep.destroy()
+        self.separators.clear()
+
+        row_num = 0
         for pack in self.packs:
             display = True
-            checkbox_var, checkbox = self.checkboxes[pack]
+            checkbox_var, checkbox, buttons = self.checkboxes[pack]
             checkbox_val = checkbox_var.get()
+            buttons.grid_forget()
             checkbox.grid_forget()
 
+            if not (self.config.install_dir / "Mods/Packs" / str(pack)).exists():
+                buttons.folder_button["state"] = tk.DISABLED
+            else:
+                buttons.folder_button["state"] = tk.NORMAL
+
             if name and name not in pack.lower():
                 display = False
 
@@ -787,7 +1002,14 @@ class PlayTab(Tab):
                 display = False
 
             if display:
-                checkbox.grid(column=0, pady=0, padx=5, sticky="w")
+                if row_num > 0:
+                    sep = ttk.Separator(self.packs_frame.scrollable_frame)
+                    sep.grid(row=row_num, column=0, pady=1, sticky="ew")
+                    self.separators.append(sep)
+                    row_num += 1
+                checkbox.grid(row=row_num, column=0, pady=0, padx=5, sticky="nsw")
+                buttons.grid(row=row_num, column=1, pady=0, padx=(5, 25), sticky="e")
+                row_num += 1
 
     def on_check(self, name, var):
         if var.get():
@@ -799,6 +1021,56 @@ class PlayTab(Tab):
     def on_check_wrapper(self, name, var):
         return lambda: self.on_check(name, var)
 
+    def open_pack_dir(self, pack):
+        if pack.startswith("/"):
+            logger.warning("Got dangerous pack name, aborting...")
+            return
+
+        pack_dir = self.config.install_dir / "Mods/Packs" / pack
+        if not pack_dir.exists():
+            logger.info("No pack directory found to remove. Looked in %s", pack_dir)
+            return
+
+        webbrowser.open(f"file://{pack_dir}")
+
+    def remove_pack(self, pack):
+        if pack.startswith("/"):
+            logger.warning("Got dangerous pack name, aborting...")
+            return
+
+        to_remove = []
+        pack_dir = self.config.install_dir / "Mods/Packs" / pack
+        if pack_dir.exists():
+            to_remove.append(pack_dir)
+
+        if pack_dir.with_suffix(".zip").exists():
+            to_remove.append(pack_dir.with_suffix(".zip"))
+
+        if not to_remove:
+            logger.info("No pack directory found to remove. Looked in %s", pack_dir)
+
+        removing = "\n".join(map(str, to_remove))
+        answer = tk.messagebox.askokcancel(
+            title="Confirmation",
+            message=(
+                "Are you sure you want to remove this pack?\n"
+                "\n"
+                "This will remove the following:\n"
+                f"{removing}"
+            ),
+            icon=tk.messagebox.WARNING,
+        )
+
+        if not answer:
+            return
+
+        for path in to_remove:
+            if path.is_dir():
+                shutil.rmtree(path)
+            else:
+                path.unlink()
+        self.on_load()
+
     def on_load(self):
         self.make_dirs()
         packs = sorted(self.get_packs())
@@ -809,20 +1081,37 @@ class PlayTab(Tab):
 
             item = tk.Checkbutton(
                 self.packs_frame.scrollable_frame,
-                text=f" {pack}",
-                image=self.default_icon,
+                text=f"{pack}",
                 font=("Segoe UI", 12, "bold"),
                 variable=var,
                 onvalue=True,
                 offvalue=False,
                 compound="left",
+                # TODO: dynamic sizing for larger windows
+                wraplength="640",
                 command=self.on_check_wrapper(pack, var),
             )
-            self.checkboxes[pack] = (var, item)
+
+            buttons = tk.Frame(self.packs_frame.scrollable_frame)
+            buttons.rowconfigure(0, weight=1)
+            buttons.folder_button = ttk.Button(
+                buttons,
+                image=self.folder_icon,
+                command=(lambda pack: lambda: self.open_pack_dir(str(pack)))(pack),
+            )
+            buttons.folder_button.grid(row=0, column=0, sticky="e")
+            buttons.trash_button = ttk.Button(
+                buttons,
+                image=self.trash_icon,
+                command=(lambda pack: lambda: self.remove_pack(str(pack)))(pack),
+            )
+            buttons.trash_button.grid(row=0, column=1, sticky="e")
+            self.checkboxes[pack] = (var, item, buttons)
 
         for pack in packs_removed:
-            (_, item) = self.checkboxes[pack]
+            (_, item, buttons) = self.checkboxes[pack]
             item.destroy()
+            buttons.destroy()
             del self.checkboxes[pack]
 
         self.packs = packs
